/*--------------------------------------------------------------.
| qxltool.c : Access QXL.WIN files from other operating systems |
|                                                               |
| (c) Jonathan Hudson, April 1998                               |
| No Warranty                                                   |
`--------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#define  __USE_GNU 1
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <signal.h>
#include <sys/stat.h>
#define _GNU_SOURCE
#include <fnmatch.h>
#include "qxltool.h"


static int help(QXL *, short mflag, char **av);
static QLDIR * readqldir(QXL *, QLDIR *cur);
static u_short nextcluster(QXL *, u_short , short);
static void updatecluster(QXL *, u_short, u_short);

u_short inline swapword (u_short val)
{
    return (u_short) (val << 8) + (val >> 8);
}

u_long inline swaplong (u_long val)
{
    return (u_long) (((u_long) swapword (val & 0xFFFF) << 16) |
		    (u_long) swapword (val >> 16));
}

#define SW(a,c) (a)->c = swapword((a)->c)
#define SL(a,c) (a)->c = swaplong((a)->c)

    
static char *strim (char *s)
{
    char *p;
    short n;

    n = strlen(s);
    p = s + n;
    while(*--p <= ' ' && n--)
    {
        *p = '\0';
    }
    return s;
}

static void h_normalise (HEADER *h)
{
    SW(h,namlen);
    SW(h,rand);
    SW(h,access);
    SW(h,sectc);
    SW(h,total);
    SW(h,free);
    SW(h,direct);    
    SW(h,formatted);
    SW(h,first);
    SL(h,dlen);
    SW(h,map[0]);
}

static void writeheader(QXL *qxl)
{
    HEADER h;
    
    lseek(qxl->fd, 0, SEEK_SET);
    qxl->h.access++;
    h = qxl->h;
    h.sectc >>= 9;
    h_normalise (&h);
    write (qxl->fd, &h, sizeof(HEADER));
}

static void readheader(QXL *qxl)
{
    lseek(qxl->fd, 0, SEEK_SET);
    read(qxl->fd, &qxl->h, sizeof(HEADER));
    h_normalise (&qxl->h);
    if(memcmp(qxl->h.id, "QLWA", 4) == 0)
    {
        // Now create a fake directory record for the root directory
        qxl->root.length = qxl->h.dlen;
        qxl->root.map = qxl->h.direct;
        qxl->root.type = 0xff;
        *qxl->root.name = 0;
        qxl->root.nlen = 0;
        qxl->h.sectc <<= 9;
        strim(qxl->h.name);
    }
    else
    {
        fputs("Not a QXL file\n", stderr);
        exit(0);
    }
}

static void d_normalise (QLDIR *d)
{
    SL(d, length);
    SW(d, type);
    SL(d, data);
    SW(d, nlen);
    SL(d, date);
    SW(d, map);
}

static int d_list(QXL *qxl, QLDIR *d, char *a, u_short b, u_short c)
{
    struct tm *tm;
    char s[128];
    size_t n;
    char *p;
    int flen,m;

    flen = d->length - sizeof(QLDIR);

    if(flen > 0 || b || (flen == 0 && d->type == 0xff))
    {
        p = d->name;
        m = d->nlen;

        if(qxl->curd.nlen)
        {
            p += qxl->curd.nlen + 1;
            m -= (qxl->curd.nlen + 1);
        }

        n = sprintf(s,"%-36.*s %8d ", m, p, flen);

        if(d->type != 0xff)
        {
            size_t k;
            
            tm = localtime(&d->date);
#ifdef __GO32__
            k = sprintf(s+n,"%02d-%02d-%02d %02d:%02d:%02d ",
                        tm->tm_mday, tm->tm_mon+1, tm->tm_year,
                        tm->tm_hour, tm->tm_min, tm->tm_sec);
#else
            k = strftime(s+n, sizeof(s)-n,
                         "%02d-%02m-%02y %02H:%02M:%02S ", tm);
#endif
            n += k;
        }
        else
        {
            strcpy (s+n, "      (Directory) ");
            n += 18;
        }

   
        if(d->type == 1)
        {
            sprintf(s+n, "(%ld) ", d->data);
        }

        if(b && d->map && flen <= 0)
        {
            u_short c;
            int k = 0;
            for(c = d->map; c; c = nextcluster(qxl, c, 0))
            {
                k++;
            }
            sprintf(s+n, "<%d Kb, 0x%x>", k*2, d->map);
        }
        
        fputs(s, qxl->fp);
        fputc('\n', qxl->fp);
    }
    return 0;
    
}

/* quick and dirty version for now */

#ifdef __GO32__
# define FFLAG 0
#else
# define FFLAG FNM_CASEFOLD
#endif

static int qfindwild(QXL *qxl, QLDIR *d, char *fn, u_short ft, u_short flag)
{
    QLDIR *pd = NULL;
    if(ft == 0 || (ft == d->type))
    {
        char fnm[40];

        strncpy(fnm, d->name, d->nlen);
        *(fnm+d->nlen) = 0;
        
        if(0 == fnmatch(fn, fnm, FFLAG))
        {
            pd = malloc(sizeof(QLDIR));
            *pd = *d;
        }
    }
    return (int)pd;
}

#undef FFLAG

static int qfindbest(QXL *qxl, QLDIR *d, char *fn, u_short ft, u_short flag)
{
    size_t wlen = strlen(fn);

    if((ft == d->type))    
    {
        if((d->nlen > qxl->lastd.nlen) && (d->nlen <= wlen) &&
           strncasecmp(d->name,fn, d->nlen) == 0)
        {
            memcpy(&qxl->lastd, d, sizeof(QLDIR));
        }
    }
    return 0;
}

static int qmatch(QXL *qxl, QLDIR *d, char *fn, u_short ft, u_short flag)
{
    QLDIR *pd = NULL;
    size_t wlen = strlen(fn);

    if(ft == 0 || (ft == d->type))
    {
        if (d->nlen == wlen && strncasecmp(d->name,fn, wlen) == 0)
        {
            pd = malloc(sizeof(QLDIR));
            *pd = *d;
        }
    }
    return (int)pd;
}

static int processdir(QXL *qxl, QLDIR *buf, QLDIR *cur,
               PCALLBACK func, char *fn, u_short ft, short flag)
{
    QLDIR *d, *last, *z;
    int res = 0;
    last = buf + cur->length / sizeof(QLDIR);
    
    for(d = buf+1; d < last; d++)
    {
        d_normalise (d);
        d->date += QL2UNIX;
        if((res = func(qxl, d, fn, ft, flag)) == 0)
        {
            if(d->type == 0xff && (flag & DO_RECURSE))
            {
                z = readqldir(qxl, d);
                res = processdir(qxl, z, d, func, fn, ft, flag);
                free(z);
            }
       
        }
        else qxl->lastd = *cur;
        if(res) break;
    }
    return res;
}

static u_short getcluster(QXL *qxl, u_short o)
{
    u_short c,n;
    qxl->h.free--;
    c = nextcluster(qxl, qxl->h.first, 0);
    n = qxl->h.first;
    if(o) updatecluster(qxl, o, n);
    updatecluster(qxl, n, 0);
    qxl->h.first = c;
    return n;
}

static u_short nextcluster(QXL *qxl, u_short c, short flag)
{
    u_long newoff;
    u_short d;
    
    newoff = c * 2 + offsetof(HEADER,map);
    lseek(qxl->fd, newoff, SEEK_SET);
    read(qxl->fd, &d, sizeof(short));
    d =swapword(d);
    if(d == 0 && flag)
    {
        if(qxl->h.free)
        {
            d = getcluster(qxl, c);
        }
        else
        {
            fputs("Disk is full!\n", stderr);
            exit(0);
        }
    }
    return d;
}

static QLDIR * readqldir(QXL *qxl, QLDIR *cur)
{
    u_char *buf,*p;
    int dlen, rlen;
    u_short fcluster;
    
    dlen = cur->length;
    p = buf = malloc(dlen);
    
    for(fcluster = cur->map; (fcluster && dlen);)
    {
        lseek(qxl->fd, fcluster*qxl->h.sectc, SEEK_SET);
        if(dlen < qxl->h.sectc)
            rlen = dlen;
        else
            rlen = qxl->h.sectc;
        read(qxl->fd, p, rlen);
        p += rlen;
        dlen -= rlen;
        fcluster = nextcluster(qxl, fcluster, 0);
    }
    return (QLDIR *)buf;
}

static void writeqldir(QXL *qxl, QLDIR *cur, QLDIR *buf)
{
    u_char *p;
    int dlen, rlen;
    u_short clu;
    u_short ifree;
    
    dlen = cur->length;
    p = (u_char *)buf;
    ifree = qxl->h.free;
    
    for(clu = cur->map; dlen; )
    {
        lseek(qxl->fd, clu*qxl->h.sectc, SEEK_SET);
        if(dlen < qxl->h.sectc)
            rlen = dlen;
        else
            rlen = qxl->h.sectc;
        write (qxl->fd, p, rlen);
        p += rlen;
        dlen -= rlen;
        if(dlen) clu = nextcluster(qxl, clu, 1);
        
    }
    if(ifree != qxl->h.free)
        writeheader(qxl);
}

static int openqxl(QXL *qxl, char *fn)
{
    if((qxl->fd = open(fn, O_RDWR|O_BINARY)) != -1)
    {
        readheader(qxl);
        memcpy(&qxl->curd, &qxl->root, sizeof(QLDIR));
        strcpy(qxl->fn, fn);
    }
    return qxl->fd;
}

static QLDIR * find_file(QXL *qxl, char *fn, u_short ft, QLDIR *cur,
                         short flag)
{
    int res = 0;
    void *buf;
    char wanted[40];
    int n;
    PCALLBACK cb;
    
    if(( n = cur->nlen))
    {
        strncpy(wanted, cur->name, n);
        *(wanted+n) = '_';
        n++;
    }
    strcpy(wanted+n,fn);

    cb = (flag & DO_BEST) ? qfindbest: (flag & DO_WILD) ? qfindwild : qmatch;
        
    buf = readqldir(qxl, cur);
    res = processdir(qxl, buf, cur, cb, wanted, ft, flag);
    free(buf);
    return (QLDIR *)res;
}


static int quit(QXL *qxl, short mflag, char **av)
{
    exit(0);
    return 0;
}

static int qcd(QXL *qxl, short mflag, char **av)
{
    QLDIR *d;
    char *p = (av) ? *av : NULL;
    
    if(p == NULL)
    {
        memcpy(&qxl->curd, &qxl->root, sizeof(QLDIR));
    }
    else
    {
        if(*(p+strlen(p)-1) == '_')
        {
            *(p+strlen(p)-1) = 0;
        }
        
        d = find_file(qxl, p, 0xff, &qxl->curd, DO_RECURSE);
        if(d)
        {
            qxl->curd = *d;
            free(d);
        }
        else fputs("Not found\n", stderr);
    }
    return 0;
}

static int qlslr (QXL *qxl, short mflag, char **av)
{
    QLDIR *buf;
    char *cmd = *av;
    
    if(cmd && *cmd)
    {
        fputs("ls[lr] takes no parameters (currently), cd first\n", stderr);
    }
    else
    {
        buf = readqldir(qxl, &qxl->curd);
        processdir(qxl, buf, &qxl->curd, d_list, NULL, 0, 1);
        free(buf);
    }
    return 0;
}

static int qls(QXL *qxl, short mflag, char **av)
{
    QLDIR *buf;
    char *cmd = *av;
    
    if(cmd && *cmd)
    {
        fputs("ls[lr] takes no parameters (currently), cd first\n", stderr);
    }
    else
    {
        buf = readqldir(qxl, &qxl->curd);
        processdir(qxl, buf, &qxl->curd, d_list, NULL, 0, 0);
        free(buf);
    }
    return 0;
}

static int qdump (QXL *qxl, short mflag, char **av)
{
    QLDIR *buf;
    char *cmd = *av;
    
    if(cmd && *cmd)
    {
        fputs("ls[lr] takes no parameters (currently), cd first\n", stderr);
    }
    else
    {
        buf = readqldir(qxl, &qxl->curd);
        processdir(qxl, buf, &qxl->curd, d_list, NULL, 1, 1);
        free(buf);
    }
    return 0;    
}


static int qpwd(QXL *qxl, short mflag, char **av)
{
    fprintf(qxl->fp, "cwd is \"%-.*s\"\n", qxl->curd.nlen, qxl->curd.name);
    return 0;
}


static void cpfile(QXL *qxl, QLDIR *d)
{
    u_char *buf;
    int dlen, rlen;
    u_short cluster;
    short off = sizeof(QLDIR);
    
    buf = alloca(qxl->h.sectc);
    dlen = d->length;
    cluster = d->map;

    if(qxl->fp)
    {
        for(cluster = d->map; cluster; cluster = nextcluster(qxl, cluster, 0))
        {
            lseek(qxl->fd, cluster*qxl->h.sectc, SEEK_SET);
            if(dlen < qxl->h.sectc)
                rlen = dlen;
            else
                rlen = qxl->h.sectc;
            read(qxl->fd, buf, rlen);
            
            fwrite(buf+off, 1, rlen-off, qxl->fp);
  
            dlen -= rlen;
            off = 0;
        }
    }

    if(d->type == 1 && qxl->close == 1 && d->data)
    {
        XTCC xtcc;
        
        fseek(qxl->fp, -8, SEEK_CUR);
        fread(&xtcc, 1, 8, qxl->fp);
        if(memcmp(xtcc.x.xtcc,"XTcc", 4))
        {
            memcpy(xtcc.x.xtcc,"XTcc", 4);
            xtcc.dlen = swaplong(d->data);
            fwrite(&xtcc, 1, 8, qxl->fp);
        }
    }
}


static int qcp(QXL *qxl, short mflag, char **av)
{
    QLDIR *d;
    char *p = *av;
    
    if(p)
    {
        d = find_file(qxl, p, 0, &qxl->curd, DO_RECURSE);
        if(d)
        {
            cpfile(qxl, d);
            free(d);
        }
        else fputs("Not found\n", stderr);
    }
    return p == NULL;
}

static void updatecluster(QXL *qxl, u_short f, u_short c)
{
    lseek(qxl->fd, f*2 + sizeof(QLDIR), SEEK_SET);
    c = swapword(c);
    write(qxl->fd, &c, sizeof(short));
}

static void updatefree(QXL *qxl, QLDIR *d)
{
    u_short c,cn,f;
   
    for(c = d->map; c; )
    {
        cn = nextcluster(qxl, c, 0);
        f = qxl->h.first;
        qxl->h.first = c;
        updatecluster(qxl, c, f);
        c = cn;
        qxl->h.free++;
    }
    writeheader(qxl);
}

static void updatedir(QXL *qxl, QLDIR *d)
{
    QLDIR *e, *f;

    e = readqldir(qxl, &qxl->lastd);
    for(f = e+1; f < (e + qxl->lastd.length/sizeof(QLDIR)); f++)
    {
        if(d->map == swapword(f->map))
        {
            memset(f, 0, sizeof(QLDIR));
            break;
        }
    }
    writeqldir(qxl, &qxl->lastd, e);
    free(e);
}

static void rmfile(QXL *qxl, QLDIR *d)
{
    char fn[40];
    memcpy(fn, d->name, d->nlen);
    *(fn+d->nlen) = 0;
    updatefree(qxl, d);
    updatedir(qxl, d);
}

static int qrm(QXL *qxl, short flag, char **av)
{
    QLDIR *d;
    char *p = *av;

    if(p)
    {
        d = find_file(qxl, p, 0, &qxl->curd, DO_RECURSE);
        if(d)
        {
            if(d->type != 0xff)
            {
                rmfile(qxl, d);
                free(d);
            }
            else fputs("Use rmdir to delete empty directories\n", stderr);
        }
        else fputs("Not found\n", stderr);
    }
    return p == NULL;
}

static int dircount(QXL *qxl, QLDIR *d)
{
    QLDIR *e,*f;
    int n = 0;
    
    e = readqldir(qxl, d);
    for(f = e+1; f < (e + d->length/sizeof(QLDIR)); f++)
    {
        if(f->nlen && swaplong(f->length) > sizeof(QLDIR))
        {
            n++;
        }
    }
    free(e);
    return n;
}


static int qrmdir(QXL *qxl, short mflag, char **av)
{
    QLDIR *d;
    char *p = *av;
    
    if(p)
    {
        d = find_file(qxl, p, 0xff, &qxl->curd, DO_RECURSE);
        if(d)
        {
            int dn;
            
            if((dn = dircount(qxl, d)) == 0)
            {
                rmfile(qxl, d);
                free(d);
            }
            else
            {
                fprintf (stderr,
                         "Directory must be empty to delete (%d)\n", dn);
            }
        }
        else fputs("Not found\n", stderr);
    }
    return p == NULL;
}

static char ** openout(QXL *qxl, char *s, short flag)
{

    char **av,**pav;
    char *r0, *r1,*r2,*p;
    int l,n,nl;
    l = strlen(s);

    pav = av = (char**)malloc(sizeof(char *)*256);
    
    if((r1 = r0 = strpbrk(s, ">|")))
    {
        r1++;
        if(*r1 == '>') r1++;
        while(*r1 <= ' ') r1++;
        r2 = r1;
        while(*r2 > ' ') r2++;
        n = r2 - r1;
        p = alloca(n+1);
        memcpy(p, r1, n);
        *(p+n) = 0;

        if(*r0 == '|')
        {
            qxl->fp = popen(p, "wb");
            qxl->close = 2;
        }
        else if(*r0 == '>')
        {
            char *mode;
            if(*(r0+1) ==  '>')
            {
                mode = (flag & QX_TEXT) ? "a" : "ab";
            }
            else mode = (flag & QX_TEXT) ? "w+" : "w+b";
            qxl->fp  = fopen (p, mode);
            qxl->close = 1;
        }
        if((nl = l - (r2 -s)) > 0)
        {
            memmove(r0, r2, nl);
        }
        *(r0+nl) = 0;
    }

    if(qxl->fp == NULL)
    {
        qxl->fp = stdout;
        qxl->close = 0;
    }

    for(n = 0, r0 = s; (*av = strtok (r0, "\t ")); r0 = 0, n++,av++)
        ;
    pav = (char **)realloc(pav, sizeof(char *)*(n+1));
    return pav;
}

static void closeout(QXL *qxl, char **pav)
{
    if(pav) free(pav);
    
    if(qxl->close == 1)
    {
        fclose(qxl->fp);
    }
    else if(qxl->close == 2)
    {
        pclose(qxl->fp);
    }
    qxl->fp = NULL;
}

static double unify(QXL *qxl, u_short x, char *f)
{
    double ts;
    
    ts = x * qxl->h.sectc;
    if(ts > 1024*1024)
    {
        ts /= 1024*1024;
        *f = 'M';
    }
    else if(ts > 1024)
    {
        ts /= 1024;
        *f = 'K';
    }
    else *f = ' ';
    return ts;
}


static int qinfo(QXL *qxl, short mflag, char **p)
{
    double ts,fs;
    char tsb, fsb;
    struct stat st;

    fstat(qxl->fd, &st);
    
    ts = unify(qxl, qxl->h.total, &tsb);
    fs = unify(qxl, qxl->h.free, &fsb);

    fprintf(qxl->fp,"Info for QXL file %s\n", qxl->fn);
    fprintf(qxl->fp,"Label %-.*s", qxl->h.namlen, qxl->h.name);
    if(strlen(qxl->h.name) != qxl->h.namlen)
    {
        fprintf(qxl->fp," [%s]", qxl->h.name);
    }
    
    fprintf(qxl->fp,"\nSector size %d bytes\n", qxl->h.sectc);    
    fprintf(qxl->fp,"Total sectors %d (%.2f %cb, on disk %d b)\n",
            qxl->h.total, ts, tsb, (int)st.st_size);
    fprintf(qxl->fp,"Free  sectors %d (%.2f %cb)\n",
            qxl->h.free, fs, fsb);
    fprintf(qxl->fp, "Formatted %d, rand %d, access %d\n",
            qxl->h.formatted, qxl->h.rand, qxl->h.access);
    return 0;
}

static void makenewfile(QXL *qxl, char *fn1, QLDIR *ld, FILE *fp, time_t ftim)
{
                
    QLDIR *e,*f;
    short got = 0;
    u_long ddata;
    u_char *buf;
    int n,m,nt;
    u_short c;
    
    e = readqldir(qxl, ld);
                    
    for(f = e+1; f < e + ld->length/sizeof(QLDIR); f++)
    {
        if(swaplong(f->length) <= sizeof(QLDIR) &&
           swapword(f->type) != 0xff)
        {
            got = 1;
            break;
        }
    }
                    
    if(got == 0)
    {
        e = realloc(e, ld->length+sizeof(QLDIR));
        f = e + ld->length/sizeof(QLDIR);
        ld->length += sizeof(QLDIR);
    }
                    
    buf = alloca(qxl->h.sectc);
        
    memset(buf, 0, qxl->h.sectc);
    memset(f, 0, sizeof(QLDIR));

    nt = 0;
    ddata = 0;
    
    if(fp)
    {
        XTCC xtcc;
        int ilen,off,nr;

        fseek(fp, -8, SEEK_END);
        fread(&xtcc, 1, 8, fp);
        if(memcmp(xtcc.x.xtcc,"XTcc", 4) == 0)
        {
            ddata = swaplong(xtcc.dlen);
        }            
        rewind(fp);

        for(nr = -1, off=sizeof(QLDIR), c=0; nr != 0; )
        {
            ilen = qxl->h.sectc-off;
            if((nr = fread(buf+off, 1, ilen, fp)) > 0)
            {
                c = getcluster(qxl, c);
                if(off)
                {
                    f->map = c;
                }
                lseek(qxl->fd, c*qxl->h.sectc, SEEK_SET);
                write(qxl->fd, buf, nr + off);
                off = 0;
                nt += nr;
            }
        }
        if(ddata)
        {
            f->type = 1;
        }
    }
    else
    {
        f->date = 0;
        f->map = c = getcluster(qxl, 0);                
        lseek(qxl->fd, c*qxl->h.sectc, SEEK_SET);
        write(qxl->fd, buf, qxl->h.sectc);
        f->type = 0xff;
    }

    f->date = ftim;
    f->length = nt+sizeof(QLDIR);
    f->data = ddata;
    
    m = strlen(fn1);
    n = 0;
        
    if(ld->map == qxl->curd.map)
    {
        if(( n = ld->nlen))
        {
            strncpy(f->name, ld->name, n);
            *(f->name+n) = '_';
            n++;
        }
    }
        
    if(m +n > 36)
    {
        m = 36-n;
    }
        
    strncpy(f->name+n,fn1, m);
    f->nlen = m + n;
    d_normalise(f);
    writeqldir(qxl, ld, e);
    free(e);
    
    if(ld->map == qxl->curd.map)
    {
        qxl->curd = *ld;
    }
    
    /* Must also update the parent */
    if(got == 0)
    {
        char fnm[40];
        QLDIR *dd;
        
        strncpy(fnm, ld->name, ld->nlen);
        *(fnm+ld->nlen) = 0;
        if((dd = find_file(qxl, fnm, 0xff, &qxl->root, DO_RECURSE)))
        {
            e = readqldir(qxl, &qxl->lastd);
            for(f = e+1; f < (e + qxl->lastd.length/sizeof(QLDIR)); f++)
            {
                if(dd->map == swapword(f->map))
                {
                    f->length = swaplong(ld->length);
                        break;
                }
            }
            writeqldir(qxl, &qxl->lastd, e);
            free(e);
            free(dd);
        }
        else
        {
            qxl->root.length += sizeof(QLDIR);
            qxl->h.dlen = qxl->root.length;
        }
    }
    writeheader(qxl);
}

static int qmkdir(QXL *qxl, short mflag, char **av)
{
    char *p = *av;
    QLDIR ld, *d = NULL;
    char *star;

    if(p)
    {
        char wild[42];
        p = stpcpy(wild, p);
        if(*(p-1) == '_')
        {
            p--;
            *p = 0;
        }
        
        if((d = find_file(qxl, wild, 0, &qxl->curd, DO_RECURSE)) == NULL)
        {
            star = p;
            *p++ = '_';
            *p++ = '*';
            *p = 0;
            d = find_file(qxl, wild, 0, &qxl->curd, DO_RECURSE|DO_WILD);
            *star = 0;
        }

        if(d)
        {
            printf("Cannot mkdir for %s as %-.*s exists\n",
                   wild, d->nlen, d->name);
            free(d);
        }
        else
        {
            qxl->lastd = qxl->root;
            find_file(qxl, wild, 0xff, &qxl->curd, DO_RECURSE|DO_BEST);
            ld = qxl->lastd;
            if(ld.nlen == 0)
            {
                ld = qxl->curd;
            }
            makenewfile(qxl, wild, &ld, NULL, 0);
        }
    }
    return p == NULL;
}

static int qwrite(QXL *qxl, short mflag, char **av)
{
    FILE *fp;
    QLDIR ld, dl, *d;
    long dlen = 0;
    char *fn0, *fn1;

    if((fn0 = *av) != NULL)
    {
        fn1 = *(av + 1);

        if(fn1 == NULL)
        {
            char *p;
            p = strrchr(fn0, '/');
            if(p == NULL)
            {
                p = strrchr(fn0, '\\');
            }
            if(p)
                p++;
            else
                p = fn0;
            fn1 = strdup(p);
            for(p = fn1; *p; p++)
            {
                if(*p == '.') *p = '_';
            }
        }
        
        if((d = find_file(qxl, fn1, 0, &qxl->curd, DO_RECURSE)) != NULL)
        {
            if(d->type != 0xff)
            {
                dlen = d->length;
                dl = qxl->lastd; 
            }
            else
            {
                dlen = -1;
                fprintf(stderr, "Can't overwrite directory \"%s\"\n", fn0);
            }
        }
        
        if(dlen >= 0)
        {
            char *mode;
            
            qxl->lastd = qxl->root;
            find_file(qxl, fn1, 0xff, &qxl->curd, DO_RECURSE|DO_BEST);
            ld = qxl->lastd;
            if(ld.nlen == 0)
            {
                ld = qxl->curd;
            }
            mode = (mflag & QX_TEXT) ? "r" : "rb";
            
            if((fp = fopen(fn0, mode)) != NULL)
            {
                struct stat st;
                fstat(fileno(fp), &st);
                if(((sizeof(QLDIR)+st.st_size) - dlen) / qxl->h.sectc >
                   qxl->h.free-1)
                {
                    fprintf(stderr,
                            "No room (%d free sectors, %ld b wanted)\n",
                            qxl->h.free, (long)st.st_size+sizeof(QLDIR));
                }
                else
                {
                    if(d)
                    {
                        qxl->lastd = dl;
                        rmfile(qxl, d);
                        free(d);
                    }
                    makenewfile(qxl, fn1, &ld, fp, st.st_ctime + UNIX2QL);
                }
                fclose(fp);
            }
            else perror(fn0);
        }
        if(*(av+1) == NULL) free(fn1);
        if(d) free(d);
    }
    return fn0 == NULL;
}


static int version(QXL *qxl, short mflag, char **av)
{
    fputs ("qxltool " VERSTR " for " TARGET "\n", stderr);
    return 0;
}


static char *defargs(char *av)
{
    char *fn;
    if(av)
    {
        fn = av;
    }
    else if((fn = getenv("QXL_FILE")) == NULL)
    {
        fn = DEFQXL;        
    }
    return fn;
}

static int qopen(QXL *qxl, short mflag, char **av)
{
    char *fn = defargs(*av);
    if(qxl->fd <= 0)
    {
        openqxl(qxl,fn);
        return 0;
    }
    else return 1;
}

static int qclose(QXL *qxl, short mflag, char **av)
{
    if(qxl->fd > 0)
    {
        close(qxl->fd);
        qxl->fd = -1;
        *qxl->fn = 0;
        return 0;
    }
    return 1;
}

static void concatargs(char **av, char *cmd)
{
    char *c = cmd;
    
    if(*av)
    {
        while(*av)
        {
            c = stpcpy(c, *av);
            *c++ = ' ';
            av++;
        }
        --c;
    }
    *c = 0;
}

static void init(QXL *qxl, char *lab, u_short bs, u_short ts)
{
    char *l;
    short ns;
    u_short i,j,k,m;
    size_t n;

    l = alloca(bs);
    
    memset(&qxl->h, 0, sizeof(HEADER));

    qxl->h.total = ts;
    qxl->h.sectc = bs;
    qxl->h.formatted = 7;
    
    n = strlen(lab);
    if(n > 20) n = 20;

    memcpy(qxl->h.name, lab, n);
    qxl->h.namlen = n;
    qxl->h.access = 0;
    memcpy(qxl->h.id, "QLWA", 4);
    ns = ((sizeof(QLDIR) + qxl->h.total*2) + bs - 1) / bs;
    qxl->h.direct = ns;
    lseek(qxl->fd, qxl->h.direct*2+sizeof(QLDIR), SEEK_SET);
    i = 0;
    write(qxl->fd, &i, sizeof(short));

    ns ++;
    qxl->h.first = ns;

    for(i = qxl->h.first; i < qxl->h.total-1; )
    {
        i++;
        j = swapword(i);
        write(qxl->fd, &j, sizeof(short));
    }

    j = 0;
    write(qxl->fd, &j, sizeof(short));

    qxl->h.free = qxl->h.total - ns - 1;
    qxl->h.dlen = sizeof(QLDIR);
    srand(time(NULL));
    qxl->h.rand = rand();
    
    writeheader(qxl);

    lseek(qxl->fd, bs*qxl->h.direct, SEEK_SET);
    l = alloca(bs);
    memset(l, 0, bs);
    fputs("Formatting ....", stdout);

    k = j = qxl->h.total/20;
    m = 5;
    
    for(i = qxl->h.direct; i < qxl->h.total; i++)
    {
        if(write(qxl->fd, l, bs) != bs)
        {
            perror("Format");
            exit(0);
        }
        if(i > k)
        {
            printf(" %3d%%\b\b\b\b\b", m);
            fflush(stdout);
            k += j;
            m += 5;
        }
    }
    fputs(" Done\n", stdout);
    qxl->root.length = qxl->h.dlen;
    qxl->root.map = qxl->h.direct;
    qxl->root.type = 0xff;
    *qxl->root.name = 0;
    qxl->root.nlen = 0;
    qxl->lastd = qxl->root;
    qxl->curd = qxl->root;    
}

static int qinit (QXL *qxl, short mflag, char **av)
{
    char lab[80];
    u_short bs = qxl->h.sectc;
    u_short ts = qxl->h.total;
    concatargs(av, lab);
    init(qxl, lab, bs, ts);
    return 0;
}

#ifdef __GO32__
# define COPYCMD(a,b)   if((b = strchr(a,'\n'))) {*b = 0;} b = a;
# define READCMD(a,b,c) fputs(a,stdout),b=fgets(c,sizeof(c),stdin) 
# define FREECMD(a,b)
# define DUPCMD(a,b)  if((a = strchr(b,'\n'))) {*a = 0;} a = strdup(b);
#else
# define COPYCMD(a,b)   strcpy(a,b)
# define READCMD(a,b,c) b=readline(a)
# define FREECMD(a,b)   if(*a) add_history(a); free(b)
# define DUPCMD(a,b)    a = b
#endif


static int qformat (QXL *qxl, short mflag, char **av)
{
#ifdef __GO32__
    char cmd[256];
#endif
    char *pcmd;
    char lab[32];
    
    char *fn = NULL, *size = NULL, *label = NULL;

    if(*av)
    {
        fn = *av++;
        if(*av)
        {
            size = *av++;
            if (*av)
            {
                concatargs(av, lab);
                label = lab;
            }
        }
    }
    
    if(label == NULL || *label == 0)
    {
        if((READCMD("QXL File name : ",pcmd,cmd)) != NULL)
        {
            DUPCMD(fn, pcmd);
            if((READCMD("Size in Mb : ",pcmd,cmd)) != NULL)
            {
                DUPCMD(size, pcmd);
                if((READCMD("Label  : ",pcmd,cmd)) != NULL)
                {
                    DUPCMD(label, pcmd);
                }
            }
        }
    }
    
    if(label)
    {
        QXL *q = malloc(sizeof(QXL));
        if(q &&
            (q->fd = open(fn, O_CREAT|O_EXCL|O_BINARY|O_RDWR, 0666)) != -1)
        {
            static short mbs[] = {32, 64, 128, 256, 0};
            u_short ss,bs,ts;
            short i;
            
            ss = strtoul(size, NULL, 10);
            if(ss < 1) ss = 1;
            if(ss > 256) ss = 256;
            for(i = 0; mbs[i]; i++)
            {
                if(ss <= mbs[i]) break;
            }
            bs = (2 << i) ;
            ts = (ss << 10)/bs;
            bs <<= 10;
            init(q, label, bs, ts);
            close(q->fd);
            free(q);
        }
        else perror("New QXL");
    }
    else fputc('\n', stdout);
    if(label != lab)
    {
        if(fn) free(fn);
        if(size) free(size);
        if(label) free(label);
    }
            
    return 0;
}


static int qsystem(QXL *qxl, short mflag, char **av)
{
    char cmd[1024],*c;

    concatargs(av, cmd);
    
    if(*cmd == 0)
    {
        if((c = getenv(SHELL)))
        {
            strcpy(cmd, c);
        }
    }

    if(*cmd)
    {
        system(cmd);
    }
    return 0;
}

static JTBL jtbl[] =
{
    {"cd", qcd, QX_OPEN},
    {"lslr",qlslr, QX_TEXT|QX_OPEN},
    {"ls", qls, QX_TEXT|QX_OPEN},
    {"dump", qdump, QX_OPEN},
    {"pwd",qpwd, QX_TEXT|QX_OPEN},
    {"cat",qcp, QX_OPEN},
    {"cp", qcp, QX_OPEN},
#ifdef __GO32__
    {"tcat",qcp, QX_TEXT|QX_OPEN},        
    {"tcp", qcp, QX_TEXT|QX_OPEN},
    {"twr", qwrite, QX_TEXT|QX_OPEN},
#endif
    {"info", qinfo,QX_TEXT|QX_OPEN},
    {"quit", quit},
    {"exit", quit},    
    {"init", qinit, QX_OPEN},
    {"wr", qwrite, QX_OPEN},
    {"rmdir", qrmdir, QX_OPEN},    
    {"rm", qrm, QX_OPEN},
    {"mkdir", qmkdir, QX_OPEN},
    {"version", version},
    {"help", help},
    {"open", qopen},
    {"close", qclose,QX_OPEN},
    {"format",qformat},
    {"shell", qsystem},
    {NULL,NULL}
};

static int help(QXL *qxl, short mflag, char **av)
{
    JTBL *j;
    fputs("Commands in qxltool " VERSION "\n\n", qxl->fp);

    for(j = jtbl; j->name; j++)
    {
        fprintf(qxl->fp, "\t%s\n", j->name);
    }
    fputs("\nRead the fine manual for more information\n", qxl->fp);
    return 0;
}

int main(int ac, char **av)
{

    char *fn = defargs(*(av+1));
    QXL *qxl = malloc(sizeof(QXL));
    
    signal(SIGPIPE, SIG_IGN);

    if((openqxl(qxl, fn)) != -1)
    {
        char cmd[256];
        char *pcmd, *ptr;
        
        while((READCMD("> ",pcmd,cmd)) != NULL)
        {
            JTBL *j;
            char **args;
            
            COPYCMD(cmd, pcmd);

            if((ptr = strchr(pcmd, '#'))) *ptr = 0;

            ptr = pcmd + strspn(pcmd, "\t ");
            
            for(j = jtbl; j->name; j++)
            {
                int n = strlen(j->name);
                if(strncasecmp(ptr, j->name, n) == 0)
                {
                    strim(ptr);
                    n += strspn(ptr+n, "\t ");
                    if(qxl->fd <= 0 && j->flag & QX_OPEN)
                    {
                        fputs("That requires an open QXL_FILE\n", stderr);
                    }
                    else
                    {
                        args = openout(qxl, ptr+n, j->flag);
                        if(j->func(qxl, j->flag, args))
                        {
                            fputs("Invalid argument(s)\n", stderr);
                        }
                        closeout(qxl, args);
                    }
                    break;
                }
            }
            if(*pcmd && j->name == NULL)
            {
                fputs("Invalid command\n", stderr);
            }
            FREECMD(cmd, pcmd);
        }
        fputs("quit\n", stdout);
    }
    else perror("QXL_FILE");
    
    return 0;
}
